import pandas as pd
from itertools import permutations
import statistics
import math
import copy

output_dir = r"D:\My Documents\Working Projects\YN 2019\Cryptocurrency\Data\out_v2\\"
input_path = output_dir + r"bitstamp_10000_sample_each_perm_epoch_60.csv"
AN_list = ["", "A", "B", "C", "D", "E", "F"]
epoch = 10


def get_to_AN(perm):
    return AN_list[perm]


def get_simple_transition_scores(perm_factor, perm_list):
    factorial = math.factorial(perm_factor)
    df = pd.DataFrame(0, index=range(factorial), columns=range(factorial))
    for idx in range(1, len(perm_list), 1):
        last_perm = perm_list[idx - 1]
        cur_perm = perm_list[idx]
        df.iloc[last_perm - 1, cur_perm - 1] += 1
    return df


def get_stats(list_of_perm, perm_factor):
    freq_dict = pd.Series(list_of_perm).value_counts(normalize=True).to_dict()
    df = get_simple_transition_scores(perm_factor, list_of_perm)
    list_of_rows_file = list()
    factorial = math.factorial(perm_factor)
    zfill_size = len(str(factorial))
    freq = [0.0 for x in range(factorial)]
    sum_perm = [0.0 for x in range(factorial)]

    list_of_rows_tmp = list(df.values.tolist())
    for h in range(len(list_of_rows_tmp)):
        sum_perm[h] = math.fsum(list_of_rows_tmp[h])

    flat_list = [item for sublist in list_of_rows_tmp for item in sublist]
    flat_list_prob = list()
    for g in range(len(flat_list)):
        g_val = flat_list[g]
        g_6_val = sum_perm[int(g / factorial)]
        if g_6_val > 0:
            res = float(g_val / g_6_val)
            flat_list_prob.append(res)
        else:
            flat_list_prob.append(0.0)

    row = list()
    sss = freq_dict.keys()
    for k in sss:
        freq[k - 1] = float(freq_dict[k])

    row.extend(freq)

    row.extend(flat_list_prob)
    list_of_rows_file.append(row)

    list_col_names_file = list()
    for k in range(len(freq)):
        list_col_names_file.append("perm_factor_" + str(perm_factor) + "_freq_" + str(k + 1).zfill(zfill_size))

    for i in range(1, factorial + 1, 1):
        for j in range(1, factorial + 1, 1):
            list_col_names_file.append(
                "perm_factor_" + str(perm_factor) + "_trans_" + str(i).zfill(zfill_size) + "_" + str(j).zfill(
                    zfill_size))

    df2 = pd.DataFrame(list_of_rows_file)
    df2.columns = list_col_names_file
    return df2.to_dict('records')[0]


def get_irr_score(list_of_perm, perm_factor):
    tmp_dict = dict()
    list_of_perm.reverse()
    factorial = math.factorial(perm_factor)
    freq = [0.0 for x in range(factorial)]
    freq_dict = pd.Series(list_of_perm).value_counts(normalize=False).to_dict()
    sss = freq_dict.keys()
    for k in sss:
        freq[k - 1] = float(freq_dict[k])

    wave_counters = [0.0 for x in range(factorial)]
    for idx in range(0, len(list_of_perm) - 1, 1):
        curr_perm = list_of_perm[idx] - 1
        next_perm = list_of_perm[idx + 1] - 1
        if (curr_perm < 3 and next_perm < 3) or (curr_perm >= 3 and next_perm >= 3):
            wave_counters[curr_perm] += 1

    sim_counters = [0.0 for x in range(factorial)]
    assim_counters = [0.0 for x in range(factorial)]

    for x in range(factorial):
        if freq[x] > 0:
            sim_counters[x] = float(100.0 * (float(wave_counters[x]) / freq[x]))
        else:
            sim_counters[x] = 0
    for x in range(factorial):
        assim_counters[x] = 100.0 - sim_counters[x]

    list_ratio = list()
    for x in range(factorial):
        tmp_dict["sim_" + str(x + 1)] = sim_counters[x]
        tmp_dict["assim_" + str(x + 1)] = assim_counters[x]
    for x in range(factorial):
        if float(sim_counters[x]) > 0:
            tmp_ratio = float(assim_counters[x]) / float(sim_counters[x])
            list_ratio.append(tmp_ratio)
            tmp_dict["assim_sim_ratio_" + str(x + 1)] = tmp_ratio
        else:
            tmp_dict["assim_sim_ratio_" + str(x + 1)] = 0

    tmp_dict["irr_sim"] = 0
    if len(list_ratio) > 0:
        tmp_dict["irr_sim"] = statistics.mean(list_ratio)
    return tmp_dict


def make_perm_list_same_size(pd_series, size):
    tmp_series = pd_series
    list_of_index = list(tmp_series.index)
    for idx in range(1, size, 1):
        if idx not in list_of_index:
            add_series = pd.Series([0.0])
            add_series.index = [idx]
            tmp_series = tmp_series.append(add_series)
    tmp_series = tmp_series.sort_index().to_list()
    return tmp_series


def calc_irreversibility_score(list_left, list_right):
    tmp_list = list()

    for i in range(len(list_left)):
        if float(list_left[i]) == 0.0:
            list_left[i] = 0.00001
        if float(list_right[i]) == 0.0:
            list_right[i] = 0.00001

    for idx in range(len(list_left)):
        tmp_val = float(list_left[idx]) * math.log(float(list_left[idx]) / float(list_right[idx]))
        tmp_list.append(tmp_val)

    return str(math.fsum(tmp_list))


def get_scores(permutation_list, permutation_backward_list):
    slice_perm = pd.Series(permutation_list)
    slice_perm_back = pd.Series(permutation_backward_list)
    slice_perm_value_counts = slice_perm.value_counts(normalize=True).sort_index()
    slice_perm_back_value_counts = slice_perm_back.value_counts(normalize=True).sort_index()

    slice_perm_value_counts = make_perm_list_same_size(slice_perm_value_counts, 7)
    slice_perm_back_value_counts = make_perm_list_same_size(slice_perm_back_value_counts, 7)

    slice_perm_value_counts_list = slice_perm_value_counts
    slice_perm_back_value_counts_list = slice_perm_back_value_counts

    irreversibility_score = calc_irreversibility_score(slice_perm_value_counts_list,
                                                       slice_perm_back_value_counts_list)

    return {"irreversibility_score": irreversibility_score}


def get_permutation(list_points):
    first = float(list_points[0])
    second = float(list_points[1])
    third = float(list_points[2])
    orig_points = [first, second, third]
    ordered_points = sorted([first, second, third])
    perm = list(permutations(ordered_points))
    for i in range(len(perm)):
        if list(perm[i]) == orig_points:
            return i + 1


def make_permutation_lists(list_prices):
    list_prices_forward = list_prices
    list_prices_backward = copy.deepcopy(list_prices)
    list_prices_backward.reverse()
    permutation_list = list()
    permutation_backward_list = list()

    for h in range(len(list_prices_forward) - 2):
        ############################################## forward
        first = float(list_prices_forward[h])
        second = float(list_prices_forward[h + 1])
        third = float(list_prices_forward[h + 2])
        perm = get_permutation([first, second, third])
        permutation_list.extend([perm])
        ############################################## forward

    for l in range(len(list_prices_backward) - 2):
        ############################################## backward
        first = float(list_prices_backward[l])
        second = float(list_prices_backward[l + 1])
        third = float(list_prices_backward[l + 2])
        perm = get_permutation([first, second, third])
        permutation_backward_list.extend([perm])
        ############################################## backward

    permutation_backward_list.pop(0)
    permutation_backward_list.pop(0)
    permutation_backward_list.reverse()
    permutation_list.pop(0)
    permutation_list.pop(0)

    return permutation_list, permutation_backward_list


def main():
    df = pd.read_csv(input_path)
    exp_list = list()
    for index, row in df.iterrows():
        perm = row[0]
        pre_perm = row[1]

        tmp_dict = dict()
        tmp_dict["perm"] = get_to_AN(perm)
        tmp_dict["pre_perm"] = get_to_AN(pre_perm)
        tmp_dict["target"] = "0"
        if tmp_dict["perm"] == "A":
            tmp_dict["target"] = "1"
        slice = row[2:2 + epoch]
        tmp_dict.update(get_stats(slice, 3))
        tmp_dict.update(get_irr_score(slice.to_list(), 3))
        rr = slice.to_list()
        rr.reverse()
        permutation_list, permutation_backward_list = make_permutation_lists(rr)
        dd = get_scores(permutation_list, permutation_backward_list)
        tmp_dict["irreversibility_score"] = dd["irreversibility_score"]
        exp_list.append(tmp_dict)

    df5 = pd.DataFrame.from_records(exp_list)
    ss = df5.sum(axis=0)
    for i in range(len(ss)):
        col_name = ss.index[i]
        if ss[i] == 0.0:
            del df5[col_name]
    path_finish = output_dir + "bitstamp_10000_sample_each_perm_epoch_" + str(epoch) + "_results.csv"
    df5.to_csv(path_finish, index=True, header=True, index_label="index")


if __name__ == "__main__":
    main()
    exit(0)
